/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_db_int.h"
#include "udm_url.h"
#include "udm_vars.h"
#include "udm_coords.h"
#include "udm_log.h"
#include "udm_hash.h"
#include "udm_doc.h"
#include "udm_word.h"
#include "udm_indexer.h"
#include "udm_parsehtml.h"

#ifdef HAVE_SQL


#define UDM_WRDCOORD(p,s)  ((((unsigned int)(s))<<24)+(unsigned int)(p))
#define UDM_WRDSEC(c)      (((unsigned int)(c))>>24)
#define UDM_WRDPOS(c)      (((unsigned int)(c))&0x001FFFFF)


static udm_rc_t
UdmDeleteWordsFromURLSingle(UDM_AGENT *Indexer, UDM_DB *db, urlid_t url_id)
{
  char qbuf[512];
  udm_snprintf(qbuf, sizeof(qbuf), "DELETE FROM dict WHERE url_id=%d", url_id);
  return UdmDBSQLQuery(Indexer, db, NULL, qbuf);
}


static udm_rc_t
StoreWordsMySQL(UDM_AGENT *A, UDM_DB *db, UDM_WORDLIST *Words, urlid_t url_id)
{
  udm_rc_t rc;
  size_t nstored= 0, i;

  if (!Words->nwords)
    return UDM_OK;

  for(nstored= 0; nstored < Words->nwords; )
  {
    char * qb,*qe;
    size_t step=1024;
    size_t mlen=1024;
    size_t rstored = 0;

    qb=(char*)UdmMalloc(mlen);
    strcpy(qb,"INSERT INTO dict (word,url_id,coord) VALUES ");
    qe=qb+strlen(qb);

    for(i= nstored; i< Words->nwords;i++)
    {
      size_t len=qe-qb;
      if (!Words->Word[i].coord.secno)
      { 
        nstored++; 
        continue;
      }
      rstored++;

      /* UDM_MAXWORDSIZE+100 should be enough */
      if((len + A->Conf->WordParam.max_word_len + 100) >= mlen)
      {
        mlen+=step;
        qb=(char*)UdmRealloc(qb,mlen);
        qe=qb+len;
      }

      if(i>nstored)*qe++=',';

      if (UdmSQLDBMode(db) == UDM_SQLDBMODE_SINGLE)
      {
        *qe++='(';
        *qe++='\'';
        strcpy(qe, Words->Word[i].word);
        while(*qe)qe++;
        *qe++='\'';
        *qe++=',';
        qe+=sprintf(qe,"%d,%d",url_id,
                    UDM_WRDCOORD(Words->Word[i].coord.pos,
                                 Words->Word[i].coord.secno) /*+ 
                                 Words->Word[i].seclen_marker*/);
        *qe++=')';
        *qe='\0';
      }
      if(qe>qb+UDM_MAX_MULTI_INSERT_QSIZE)
        break;
    }
    nstored = i + 1;
    rc = (rstored > 0) ? UdmDBSQLQuery(A, db, NULL, qb) : UDM_OK;
    UDM_FREE(qb);
    if(rc!=UDM_OK)
      break;
  }
  return rc;
}


static udm_rc_t
StoreWordsSingleGeneric(UDM_AGENT *A, UDM_DB *db,
                        UDM_WORDLIST *Words, urlid_t url_id)
{
  size_t i;
  for(i= 0; i < Words->nwords; i++)
  {
    udm_rc_t rc;
    char qbuf[256];
    if (!Words->Word[i].coord.secno)
      continue;

    if (UdmSQLDBMode(db) == UDM_SQLDBMODE_SINGLE)
    {
      sprintf(qbuf,"INSERT INTO dict (url_id,word,coord) VALUES(%d,'%s',%d)",
              url_id,
              Words->Word[i].word,
              UDM_WRDCOORD(Words->Word[i].coord.pos,
              Words->Word[i].coord.secno));
      if (UDM_OK!= (rc= UdmDBSQLQuery(A, db, NULL, qbuf)))
        return rc;
    }
  }
  return UDM_OK;
}


static udm_rc_t
StoreWordsSingle(UDM_AGENT *Indexer, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc= UDM_OK;
  urlid_t  url_id = UdmVarListFindInt(&Doc->Sections, "ID", 0);
  UDM_WORDLIST Words;

  UdmWordListInit(&Words);
  if (UDM_OK != (rc= UdmPrepareWords(Indexer, Doc, &Words)))
    goto ret;

  if (UDM_OK != (rc= UdmDeleteWordsFromURLSingle(Indexer, db, url_id)))
    goto ret;
  
  if (UdmVarListFindInt(&Indexer->Conf->Vars, "SaveSectionSize", 1))
  {
    if (UDM_OK!= (rc= UdmWordListSaveSectionSize(&Words)))
      goto ret;
  }
  
  /*
    Don't need to delete words here,
    they are deleted in UdmStoreWords().
  */

  /* Insert new words */
  if (UdmSQLDBType(db) == UDM_DB_MYSQL)
  {
    rc= StoreWordsMySQL(Indexer, db, &Words, url_id);
  }
  else
  {
    rc= StoreWordsSingleGeneric(Indexer, db, &Words, url_id);
  }
ret:
  UdmWordListFree(&Words);
  return rc;
}


static udm_rc_t
UdmFindWordSingleInternal(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                          UDM_FINDWORD_ARGS *args,
                          UDM_URLCRDLIST *CoordList,
                          const char *table, int join)
{
  char qbuf[4096];
  UDM_SQLRES SQLRes;
  size_t numrows, i, curnum;
  udm_rc_t rc;

  if (*args->where)
  {
    udm_snprintf(qbuf, sizeof(qbuf) - 1,"\
SELECT %s.url_id,%s.coord FROM %s, url%s \
WHERE %s.%s AND url.rec_id=%s.url_id AND %s",
    table, table, table, UdmSQLDBQueryFrom(Query),
    table, args->cmparg, table, args->where);
  }
  else if (!join)
  {
    udm_snprintf(qbuf, sizeof(qbuf) - 1,
      "SELECT url_id,coord FROM %s WHERE %s", table, args->cmparg);
  }
  else
  {

    udm_snprintf(qbuf,sizeof(qbuf)-1,"\
SELECT url_id,coord FROM %s,url WHERE %s.%s AND url.rec_id=%s.url_id",
       table, table, args->cmparg, table);
  }
  
  if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLRes, qbuf)))
    return rc;
  
  if ((numrows= UdmSQLNumRows(&SQLRes)))
  {
    CoordList->Coords= (UDM_URL_CRD*) UdmMalloc(numrows*sizeof(UDM_URL_CRD));
    CoordList->acoords= numrows;
  }
  
  /* Add new found word to the list */
  for(curnum= 0, i= 0; i < numrows; i++)
  {
    uint4 coord= atoi(UdmSQLValue(&SQLRes, i, 1));
    uint4 section= UDM_WRDSEC(coord);
    uint4 weight= args->query_param.wf[section];
    
    if(weight && (!args->Word.Param.secno || args->Word.Param.secno == section))
    {
      UDM_URL_CRD *Coord= &CoordList->Coords[curnum];
      Coord->urlid_coord.url_id= UDM_ATOI(UdmSQLValue(&SQLRes, i, 0));
      Coord->urlid_coord.coord.pos= UDM_WRDPOS(coord);
      Coord->urlid_coord.coord.secno= UDM_WRDSEC(coord);
      Coord->num= args->Word.Param.order & 0xFF;
      Coord->seclen= 0;
      curnum++;
    }
  }
  CoordList->ncoords= curnum;
  UdmSQLFree(&SQLRes);
  UdmURLCRDListSortByURLThenSecnoThenPos(CoordList);
  return rc;
}


static udm_rc_t
UdmFindWordSingle(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                  UDM_FINDWORD_ARGS *args)
{
  udm_rc_t rc;
  UDM_URLCRDLIST CoordList;

  bzero(&CoordList, sizeof(CoordList));

  if (UDM_OK != (rc= UdmFindWordSingleInternal(A, db, Query, args,
                                               &CoordList, "dict", 0)))
    return rc;

  if (args->query_param.SaveSectionSize && CoordList.ncoords)
  {
    UDM_URL_CRD *Crd= CoordList.Coords;
    UDM_URL_CRD *End= CoordList.Coords + CoordList.ncoords;
    UDM_URL_CRD *To= CoordList.Coords;
    UDM_URL_CRD *Prev= CoordList.Coords;
    urlid_t prev_url_id;
    udm_secno_t prev_secno;
    
    for (prev_url_id= Crd->urlid_coord.url_id,
         prev_secno= Crd->urlid_coord.coord.secno ;
         Crd < End; Crd++)
    {
      UDM_URL_CRD *Next= Crd + 1;
      if (Next == End ||
          Next->urlid_coord.url_id != prev_url_id ||
          Next->urlid_coord.coord.secno  != prev_secno)
      {
        for ( ; Prev < To; Prev++)
        {
          Prev->seclen= Crd->urlid_coord.coord.pos;
        }
        if (Next < End)
        {
          prev_url_id= Next->urlid_coord.url_id;
          prev_secno= Next->urlid_coord.coord.secno;
          Prev= To;
        }
      }
      else
        *To++= *Crd;
    }
    CoordList.ncoords= To - CoordList.Coords;
  }
  if (args->urls.nurls)
    UdmApplyFastLimit(&CoordList, &args->urls);
  if (CoordList.ncoords)
  {
    args->Word.Param.count= CoordList.ncoords;
    UdmURLCRDListListAddWithSort2(&args->SearchSectionListList,
                                  &Query->Res.WWList,
                                  &args->Word, &CoordList);
  }
  else
  {
    UdmFree(CoordList.Coords);
  }
  return(UDM_OK);
}


static udm_rc_t
UdmTruncateDictSingle(UDM_AGENT *A, UDM_DB *db)
{
  return UdmDBSQLTableTruncateOrDelete(A, db, "dict");
}


static udm_rc_t
UdmWordStatCreateSingle(UDM_AGENT *A, UDM_DB *db)
{
  char qbuf[128];
  sprintf(qbuf, "SELECT word, count(*) FROM dict GROUP BY word");
  return UdmWordStatQuery(A, db, qbuf);
}


static udm_rc_t
UdmDumpWordInfoOneDocSingle(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  return UDM_OK;
}


static udm_rc_t
UdmSingleInitSearch(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                    UDM_FINDWORD_ARGS *args)
{
  return UDM_OK;
}


static udm_rc_t
UdmQueryActionSingle(UDM_AGENT *A, UDM_DB *db,
                     UDM_QUERY *Query, udm_querycmd_t cmd)
{
  switch (cmd)
  {
    case UDM_QUERYCMD_CLEAR:    return UdmTruncateDictSingle(A, db);
    case UDM_QUERYCMD_WORDSTAT: return UdmWordStatCreateSingle(A, db);
    case UDM_QUERYCMD_INDEX:
      return udm_dbmode_handler_blob.QueryAction(A, db, Query, cmd);
    default: break;
  }
  return UDM_NOTARGET;
}


const UDM_DBMODE_HANDLER udm_dbmode_handler_single=
{
  "single",
  StoreWordsSingle,
  UdmQueryActionSingle,
  UdmDeleteWordsFromURLSingle,
  UdmFindWordSingle,
  UdmDumpWordInfoOneDocSingle,
  UdmSingleInitSearch,
};


#endif /* HAVE_SQL */
