/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/


#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "udm_common.h"

typedef struct
{
  UDM_VALUE generic;
  char Value;
} UDM_VALUE_CHAR;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_CHAR Value;
} UDM_VAR_CHAR;


static udm_rc_t
UdmValueCharDump(const char *Data, const char *name, FILE *f)
{
  return UDM_NOTARGET;
}


static udm_var_flag_t
UdmValueCharFlags(const char *Data)
{
  return UDM_VARFLAG_NONE;
}


static udm_secno_t
UdmValueCharSecno(const char *Data)
{
  return 0;
}


static size_t
UdmValueCharMemUsed(const char *Data)
{
  return sizeof(char);
}


static void
UdmValueCharDestructor(char *Data)
{
}


static udm_rc_t
UdmValueCharCopy(char *D, const char *S)
{
  UDM_ASSERT(0); /* TODO34: Add CloneVar(), with reference */
  UDM_BZERO(D, sizeof(char));
  return UDM_OK;
}


static udm_rc_t
UdmValueCharConstructor(char *Data, const UDM_VALUE **args, size_t nargs)
{
  *Data= 0;
  return UDM_OK;
}


static size_t
UdmValueCharVarSize(void)
{
  return sizeof(UDM_VAR_CHAR);
}


static size_t
UdmValueCharValueSize(void)
{
  return sizeof(UDM_VALUE_CHAR);
}


static size_t
UdmValueCharDataOffset(void)
{
  return offsetof(UDM_VALUE_CHAR, Value);
}


static size_t
UdmValueCharDataSize(void)
{
  return sizeof(char);
}


static size_t
UdmValueCharDataAlignment(void)
{
  return sizeof(char);
}


static void
UdmValueCharGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  static const UDM_CONST_STR name= {UDM_CSTR_WITH_LEN("#Char#")};
  *Val= name;
}


static void
UdmValueCharGetBool(const char *Data, udm_bool_t *val)
{
  *val= Data[0] != 0;
}


static void
UdmValueCharGetInt(const char *Data, int *val)
{
  *val= Data[0];
}


static void
UdmValueCharGetUInt(const char *Data, unsigned int *val)
{
  *val= Data[0];
}


static void
UdmValueCharGetDouble(const char *Data, double *val)
{
  *val= Data[0];
}


static udm_rc_t
UdmValueCharSetInt(char *Data, int value)
{
  Data[0]= value;
  return UDM_OK;
}


static udm_rc_t
UdmValueCharSetDouble(char *Data, double value)
{
  Data[0]= value;
  return UDM_OK;
}


static udm_rc_t
UdmValueCharConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueCharAppendStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueCharSetStrn(char *Data, const char *str, size_t length)
{
  char tmp[128];
  udm_snprintf(tmp, sizeof(tmp), "%.*s", (int) length, str);
  Data[0]= atoi(tmp);
  return UDM_OK;
}


static udm_rc_t
UdmValueCharSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                    const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueCharAppendConv(char *Data, size_t maxlen,
                      UDM_CONV *cnv, int cnvflags,
                      const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueCharPrintToFile(const char *Data, FILE *f)
{
  fprintf(f, "%c", Data[0]);
  return UDM_OK;
}


static udm_rc_t
UdmValueCharPrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  return UdmDSTRAppend(dstr, Data, 1);
}


UDM_VALUE_HANDLER UdmValueHandlerChar=
{
  UDM_VALUE_HANDLER_TYPE_CHAR,
  UDM_VALUE_DATA_TYPE_CHAR,
  UDM_VALUE_DATA_TYPE_INT,
  "char",
  UdmValueCharConstructor,
  UdmValueCharDestructor,
  UdmValueCharCopy,
  UdmValueCharVarSize,
  UdmValueCharValueSize,
  UdmValueCharDataOffset,
  UdmValueCharDataSize,
  UdmValueCharDataAlignment,
  UdmValueCharDump,
  UdmValueCharFlags,
  UdmValueCharSecno,
  UdmValueCharMemUsed,
  UdmValueCharGetConstStr,
  UdmValueCharGetBool,
  UdmValueCharGetInt,
  UdmValueCharGetUInt,
  UdmValueCharGetDouble,
  UdmValueCharSetInt,
  UdmValueCharSetDouble,
  UdmValueCharSetStrn,
  UdmValueCharPrintToFile,
  UdmValueCharPrintToDSTR,
  UdmValueCharConvertCharset,
  UdmValueCharSetConv,
  UdmValueCharAppendStrn,
  UdmValueCharAppendConv,
  NULL,                    /* Methods */
};
