/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/


#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"


typedef struct
{
  UDM_VALUE generic;
  double Value;
} UDM_VALUE_DOUBLE;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_DOUBLE Value;
} UDM_VAR_DOUBLE;


static size_t
UdmValueDoubleDataOffset(void)
{
  return offsetof(UDM_VALUE_DOUBLE, Value);
}


static size_t
UdmValueDoubleDataSize(void)
{
  return sizeof(double);
}


static size_t
UdmValueDoubleDataAlignment(void)
{
  return sizeof(double);
}


static udm_rc_t
UdmValueDoubleDump(const char *Data, const char *name, FILE *f)
{
  return UDM_NOTARGET;
}


static udm_var_flag_t
UdmValueDoubleFlags(const char *Data)
{
  return UDM_VARFLAG_NONE;
}


static udm_secno_t
UdmValueDoubleSecno(const char *Data)
{
  return 0;
}


static size_t
UdmValueDoubleMemUsed(const char *Data)
{
  return sizeof(double);
}


static void
UdmValueDoubleDestructor(char *Data)
{
}


static udm_rc_t
UdmValueDoubleCopy(char *D, const char *S)
{
  UDM_ASSERT(0); /* TODO34: Add CloneVar(), with reference */
  memmove(D, S, sizeof(double));
  return UDM_OK;
}


static udm_rc_t
UdmValueDoubleConstructor(char *Data,
                          const UDM_VALUE **args, size_t nargs)
{
  ((double *) Data)[0]= 0;
  return UDM_OK;
}


static size_t
UdmValueDoubleVarSize(void)
{
  return sizeof(UDM_VAR_DOUBLE);
}


static size_t
UdmValueDoubleValueSize(void)
{
  return sizeof(UDM_VALUE_DOUBLE);
}


static void
UdmValueDoubleGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  static const UDM_CONST_STR name= {UDM_CSTR_WITH_LEN("#Double#")};
  *Val= name;
}


static void
UdmValueDoubleGetBool(const char *Data, udm_bool_t *val)
{
  *val= *((const double*) Data);
}


static void
UdmValueDoubleGetInt(const char *Data, int *val)
{
  *val= *((const double*) Data);
}


static void
UdmValueDoubleGetUInt(const char *Data, unsigned int *val)
{
  *val= *((const double*) Data);
}


static void
UdmValueDoubleGetDouble(const char *Data, double *val)
{
  *val= *((const double*) Data);
}


static udm_rc_t
UdmValueDoubleSetInt(char *Data, int value)
{
  ((double *) Data)[0]= value;
  return UDM_OK;
}


static udm_rc_t
UdmValueDoubleSetDouble(char *Data, double value)
{
  ((double *) Data)[0]= value;
  return UDM_OK;
}


static udm_rc_t
UdmValueDoubleConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDoubleAppendStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDoubleSetStrn(char *Data, const char *str, size_t length)
{
  char tmp[128];
  udm_snprintf(tmp, sizeof(tmp), "%.*s", (int) length, str);
  ((double *) Data)[0]= atof(tmp);
  return UDM_OK;
}


static udm_rc_t
UdmValueDoubleSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                      const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDoubleAppendConv(char *Data, size_t maxlen,
                         UDM_CONV *cnv, int cnvflags,
                         const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueDoublePrintToFile(const char *Data, FILE *f)
{
  fprintf(f, "%f", ((const double *)Data)[0]);
  return UDM_OK;
}


static udm_rc_t
UdmValueDoublePrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  char tmp[256];
  size_t length= udm_snprintf(tmp, sizeof(tmp), "%f", ((const double *)Data)[0]);
  return UdmDSTRAppend(dstr, tmp, length);
}



UDM_VALUE_HANDLER UdmValueHandlerDouble=
{
  UDM_VALUE_HANDLER_TYPE_DOUBLE,
  UDM_VALUE_DATA_TYPE_DOUBLE,
  UDM_VALUE_DATA_TYPE_DOUBLE,
  "double",
  UdmValueDoubleConstructor,
  UdmValueDoubleDestructor,
  UdmValueDoubleCopy,
  UdmValueDoubleVarSize,
  UdmValueDoubleValueSize,
  UdmValueDoubleDataOffset,
  UdmValueDoubleDataSize,
  UdmValueDoubleDataAlignment,
  UdmValueDoubleDump,
  UdmValueDoubleFlags,
  UdmValueDoubleSecno,
  UdmValueDoubleMemUsed,
  UdmValueDoubleGetConstStr,
  UdmValueDoubleGetBool,
  UdmValueDoubleGetInt,
  UdmValueDoubleGetUInt,
  UdmValueDoubleGetDouble,
  UdmValueDoubleSetInt,
  UdmValueDoubleSetDouble,
  UdmValueDoubleSetStrn,
  UdmValueDoublePrintToFile,
  UdmValueDoublePrintToDSTR,
  UdmValueDoubleConvertCharset,
  UdmValueDoubleSetConv,
  UdmValueDoubleAppendStrn,
  UdmValueDoubleAppendConv,
  NULL,                    /* Methods */
};
