/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#define UDM_VAR_H_PRIVATE

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_env.h"
#include "udm_agent.h"
#include "udm_result.h"
#include "udm_conf.h"
#include "udm_db.h"           /* For UdmFind() */
#include "udm_sqldbms.h"
#include "udm_db_int.h"
#include "udm_searchtool.h"
#include "udm_server.h"
#include "udm_log.h"
#include "udm_prog.h"


typedef struct
{
  UDM_VALUE generic;
  UDM_ENV *Env;
} UDM_VALUE_ENV;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_ENV Value;
} UDM_VAR_ENV;


static udm_rc_t
UdmValueEnvDump(const char *Data, const char *name, FILE *f)
{
  return UDM_NOTARGET;
}


static udm_var_flag_t
UdmValueEnvFlags(const char *Data)
{
  return UDM_VARFLAG_READONLY;
}


static udm_secno_t
UdmValueEnvSecno(const char *Data)
{
  return 0;
}


static size_t
UdmValueEnvMemUsed(const char *Data)
{
  return sizeof(UDM_ENV);
}


static udm_rc_t
UdmValueEnvConstructor(char *Data,
                       const UDM_VALUE **args, size_t nargs)
{
  UDM_ENV *Env;
  if (!(Env= (UDM_ENV*) UdmMalloc(sizeof(UDM_ENV))))
    return UDM_ERROR;
  UdmEnvInit(Env);
  memmove(Data, &Env, sizeof(UDM_ENV *));
  return UDM_OK;
}


static void
UdmValueEnvDestructor(char *Data)
{
  UDM_ENV *Env;
  memmove(&Env, Data, sizeof(UDM_ENV *));
  UdmEnvFree(Env);
  UdmFree(Env);
}


static size_t
UdmValueEnvVarSize(void)
{
  return sizeof(UDM_VAR_ENV);
}


static size_t
UdmValueEnvValueSize(void)
{
  return sizeof(UDM_VALUE_ENV);
}


static size_t
UdmValueEnvDataOffset(void)
{
  return offsetof(UDM_VALUE_ENV, Env);
}


static size_t
UdmValueEnvDataSize(void)
{
  return sizeof(UDM_ENV*);
}


static size_t
UdmValueEnvDataAlignment(void)
{
  return sizeof(UDM_ENV*);
}


static udm_rc_t
UdmValueEnvCopy(char *D, const char *S)
{
  UDM_ASSERT(0); /* TODO34: Add CloneVar(), with reference */
  UDM_BZERO(D, sizeof(UDM_ENV *));
  return UDM_OK;
}


static udm_rc_t
UdmEnvAddLineConstStr(UDM_ENV *Env, UDM_CONST_STR *cstr)
{
  udm_rc_t rc;
  UDM_CFG Cfg;
  UDM_AGENT Agent;
  UDM_SERVER Srv;
  UDM_STR str;
  UdmServerInit(&Srv);
  Env->Cfg_Srv= &Srv;
  bzero((void*)&Cfg, sizeof(Cfg));
  Cfg.Indexer= &Agent;
  Cfg.Srv= &Srv;
  Cfg.flags= UDM_FLAG_SPELL;

  if (UDM_OK != UdmStrNDup(&str, cstr->str, cstr->length))
    return UDM_ERROR;
  UdmAgentInit(&Agent, Env, 0);
  rc= UdmEnvAddLine(&Cfg, str.str);
  UdmStrFree(&str);
  UdmAgentFree(&Agent);
  /*
    Some variables (e.g. DetectClones) are collected to Srv,
    not to Env directly. Copy from Srv to Env.
  */
  UdmVarListReplaceLst(&Env->Vars, &Srv.Vars, NULL, "*");
  UdmServerFree(&Srv);
  Env->Cfg_Srv= NULL;
  return rc;
}


static void
UdmEnvMethodAddLine(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_CONST_STR str;
  UdmValueHandlerSimple.GetConstStr(state->vreg[1].Data, &str);
  UDM_FUNC_RETURN_INT(state,
                      UdmEnvAddLineConstStr(((UDM_ENV **) state->vreg[0].Data)[0],
                                            &str));
}


static void
UdmEnvMethodErrMsg(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_ENV *Env= ((UDM_ENV **) state->vreg[0].Data)[0];
  const char *errmsg= UdmEnvErrMsg(Env);
  UdmProgExecutorStateReturnStrn(state, errmsg, strlen(errmsg));
}


static UDM_VALUE *
UdmVarListFindValueByConstStr(UDM_VARLIST *Vars, const UDM_CONST_STR *str)
{
  char name[128];
  udm_snprintf(name, sizeof(name), "%.*s", (int) str->length, str->str);
  return UdmVarListFindValue(Vars, name);
}


static void
UdmEnvMethodProperty(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_ENV *Env= ((UDM_ENV **) state->vreg[0].Data)[0];
  UDM_CONST_STR str;
  UDM_VALUE *Value;
  UdmValueHandlerSimple.GetConstStr(state->vreg[1].Data, &str);
  if (!(Value= UdmVarListFindValueByConstStr(&Env->Vars, &str)))
  {
    UdmProgExecutorStateReturnStrn(state, "", 0);
  }
  else
  {
    Value->handler->GetConstStr(UdmValueDataPtr(Value), &str);
    UdmProgExecutorStateReturnStrn(state, str.str, str.length);
  }
}


UDM_VALUE_PROTOTYPE UdmEnvMethodPrototypeNoArg[]=
{{&UdmValueHandlerEnv}};

UDM_VALUE_PROTOTYPE UdmEnvMethodPrototypeAddLine[]=
{{&UdmValueHandlerEnv},{&UdmValueHandlerSimple}};

UDM_VALUE_PROTOTYPE UdmEnvMethodPrototypeProperty[]=
{{&UdmValueHandlerEnv},{&UdmValueHandlerSimple}};

static const UDM_FUNCTION EnvMethods[]=
{
  {{UDM_CSTR_WITH_LEN("addline")},
    {{&UdmValueHandlerInt}, 2, UdmEnvMethodPrototypeAddLine},
    UdmEnvMethodAddLine},
  {{UDM_CSTR_WITH_LEN("errmsg")},
    {{&UdmValueHandlerSimple}, 1, UdmEnvMethodPrototypeNoArg},
    UdmEnvMethodErrMsg},
  {{UDM_CSTR_WITH_LEN("property")},
    {{&UdmValueHandlerSimple}, 2, UdmEnvMethodPrototypeProperty},
    UdmEnvMethodProperty},
  {{0,0}, {{NULL}, 0, NULL}, NULL}
};


#ifdef HAVE_SQL /*TODO34*/

#endif /* HAVE_SQL */


UDM_VALUE_PROTOTYPE UdmEnvMethodPrototypeFindOld[]=
{
 {&UdmValueHandlerEnv},     /* Env */
 {&UdmValueHandlerSimple},  /* Query */
 {&UdmValueHandlerSimple}   /* Result name */
};

#ifdef HAVE_SQL
UDM_VALUE_PROTOTYPE UdmEnvMethodPrototypeSQLQuery[]=
{
 {&UdmValueHandlerEnv},     /* Env */
 {&UdmValueHandlerSimple},  /* SQL query */
 {&UdmValueHandlerSimple}   /* Result name */
};


#endif


static void
UdmValueEnvGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  static const UDM_CONST_STR name= {UDM_CSTR_WITH_LEN("#Env#")};
  *Val= name;
}


static void
UdmValueEnvGetBool(const char *D, udm_bool_t *val)
{
  *val= UDM_TRUE;
}


static void
UdmValueEnvGetInt(const char *D, int *val)
{
  *val= 0;
}


static void
UdmValueEnvGetUInt(const char *D, unsigned int *val)
{
  *val= 0;
}


static void
UdmValueEnvGetDouble(const char *D, double *val)
{
  *val= 0.0;
}


static udm_rc_t
UdmValueEnvConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueEnvAppendStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueEnvAppendConv(char *Data, size_t maxlen,
                      UDM_CONV *cnv, int cnvflags,
                      const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueEnvSetStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueEnvSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                   const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueEnvSetInt(char *Data, int value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueEnvSetDouble(char *Data, double value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueEnvPrintToFile(const char *Data, FILE *f)
{
  UDM_CONST_STR str;
  UdmValueEnvGetConstStr(Data, &str);
  fprintf(f, "%.*s", (int) str.length, str.str);
  return UDM_OK;
}


static udm_rc_t
UdmValueEnvPrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  UDM_CONST_STR str;
  UdmValueEnvGetConstStr(Data, &str);
  UdmDSTRAppend(dstr, str.str, str.length);
  return UDM_OK;
}


UDM_VALUE_HANDLER UdmValueHandlerEnv=
{
  UDM_VALUE_HANDLER_TYPE_ENV,
  UDM_VALUE_DATA_TYPE_ENV,
  UDM_VALUE_DATA_TYPE_STR,
  "ENV",
  UdmValueEnvConstructor,
  UdmValueEnvDestructor,
  UdmValueEnvCopy,
  UdmValueEnvVarSize,
  UdmValueEnvValueSize,
  UdmValueEnvDataOffset,
  UdmValueEnvDataSize,
  UdmValueEnvDataAlignment,
  UdmValueEnvDump,
  UdmValueEnvFlags,
  UdmValueEnvSecno,
  UdmValueEnvMemUsed,
  UdmValueEnvGetConstStr,
  UdmValueEnvGetBool,
  UdmValueEnvGetInt,
  UdmValueEnvGetUInt,
  UdmValueEnvGetDouble,
  UdmValueEnvSetInt,
  UdmValueEnvSetDouble,
  UdmValueEnvSetStrn,
  UdmValueEnvPrintToFile,
  UdmValueEnvPrintToDSTR,
  UdmValueEnvConvertCharset,
  UdmValueEnvSetConv,
  UdmValueEnvAppendStrn,
  UdmValueEnvAppendConv,
  EnvMethods,
};
