/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_vars.h"
#include "udm_db.h"           /* For UdmFind() */
#include "udm_sqldbms.h"
#include "udm_db_int.h"
#include "udm_prog.h"

#ifdef HAVE_SQL /* TODO34 */


typedef struct
{
  UDM_VALUE generic;
  UDM_SQLRES Result;
} UDM_VALUE_SQLRES;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_SQLRES Value;
} UDM_VAR_SQLRES;


static udm_rc_t
UdmValueSQLResultDump(const char *Data, const char *name, FILE *f)
{
  return UDM_NOTARGET;
}


static udm_var_flag_t
UdmValueSQLResultFlags(const char *Data)
{
  return UDM_VARFLAG_READONLY;
}


static udm_secno_t
UdmValueSQLResultSecno(const char *Data)
{
  return 0;
}


static size_t
UdmValueSQLResultMemUsed(const char *Data)
{
  return sizeof(UDM_SQLRES);
}


static udm_rc_t
UdmValueSQLResultConstructor(char *Data,
                             const UDM_VALUE **args, size_t nargs)
{
  UDM_BZERO(Data, sizeof(UDM_SQLRES));
  return UDM_OK;
}


static size_t
UdmValueSQLResultVarSize(void)
{
  return sizeof(UDM_VAR_SQLRES);
}


static size_t
UdmValueSQLResultValueSize(void)
{
  return sizeof(UDM_VALUE_SQLRES);
}


static size_t
UdmValueSQLResultDataOffset(void)
{
  return offsetof(UDM_VALUE_SQLRES, Result);
}


static size_t
UdmValueSQLResultDataSize(void)
{
  return sizeof(UDM_SQLRES);
}


static size_t
UdmValueSQLResultDataAlignment(void)
{
  return sizeof(size_t);
}


static void
UdmValueSQLResultDestructor(char *Data)
{
  UdmSQLFree((UDM_SQLRES*) Data);
}


static udm_rc_t
UdmValueSQLResultCopyValue(char *D, const char *S)
{
  UDM_ASSERT(0); /* TODO34: Add CloneVar(), with reference */
  memmove(D, S, sizeof(UDM_SQLRES));
  return UDM_OK;
}


static void
UdmValueSQLResultGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  static const UDM_CONST_STR name= {UDM_CSTR_WITH_LEN("#SQLResult#")};
  *Val= name;
}


static void
UdmValueSQLResultGetBool(const char *Data, udm_bool_t *val)
{
  *val= UDM_TRUE;
}


static void
UdmValueSQLResultGetInt(const char *Data, int *val)
{
  *val= 0;
}


static void
UdmValueSQLResultGetUInt(const char *Data, unsigned int *val)
{
  *val= 0;
}


static void
UdmValueSQLResultGetDouble(const char *Data, double *val)
{
  *val= 0.0;
}


static udm_rc_t
UdmValueSQLResultSetInt(char *Data, int value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueSQLResultSetDouble(char *Data, double value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueSQLResultConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueSQLResultAppendStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueSQLResultAppendConv(char *Data, size_t maxlen,
                            UDM_CONV *cnv, int cnvflags,
                            const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueSQLResultSetStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueSQLResultSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                         const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueSQLResultPrintToFile(const char *Data, FILE *f)
{
  return UDM_OK;
}


static udm_rc_t
UdmValueSQLResultPrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  return UDM_OK;
}


UDM_VALUE_HANDLER UdmValueHandlerSQLResult=
{
  UDM_VALUE_HANDLER_TYPE_SQLRESULT,
  UDM_VALUE_DATA_TYPE_SQLRESULT,
  UDM_VALUE_DATA_TYPE_STR,
  "UdmSQLResult",
  UdmValueSQLResultConstructor,
  UdmValueSQLResultDestructor,
  UdmValueSQLResultCopyValue,
  UdmValueSQLResultVarSize,
  UdmValueSQLResultValueSize,
  UdmValueSQLResultDataOffset,
  UdmValueSQLResultDataSize,
  UdmValueSQLResultDataAlignment,
  UdmValueSQLResultDump,
  UdmValueSQLResultFlags,
  UdmValueSQLResultSecno,
  UdmValueSQLResultMemUsed,
  UdmValueSQLResultGetConstStr,
  UdmValueSQLResultGetBool,
  UdmValueSQLResultGetInt,
  UdmValueSQLResultGetUInt,
  UdmValueSQLResultGetDouble,
  UdmValueSQLResultSetInt,
  UdmValueSQLResultSetDouble,
  UdmValueSQLResultSetStrn,
  UdmValueSQLResultPrintToFile,
  UdmValueSQLResultPrintToDSTR,
  UdmValueSQLResultConvertCharset,
  UdmValueSQLResultSetConv,
  UdmValueSQLResultAppendStrn,
  UdmValueSQLResultAppendConv,
  NULL,
};
#endif /* HAVE_SQL */
