/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_vars.h"
#include "udm_log.h"
#include "udm_sgml.h"
#include "udm_word.h"

#include "udm_prog.h"


typedef struct
{
  UDM_VALUE generic;
  UDM_WIDEWORD Word;
} UDM_VALUE_QUERYWORD;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_QUERYWORD Word;
} UDM_VAR_QUERYWORD;


static udm_rc_t
UdmValueQueryWordConstructor(char *Data,
                             const UDM_VALUE **args, size_t nargs)
{
  UDM_WIDEWORD *Word= (UDM_WIDEWORD *) Data;
  UDM_ASSERT(nargs == 0);
  UdmWideWordInit(Word);
  return UDM_OK;
}


static void
UdmValueQueryWordDestructor(char *Data)
{
  UDM_WIDEWORD *Word= (UDM_WIDEWORD *) Data;
  UdmWideWordFree(Word);
}


static size_t
UdmValueQueryWordVarSize(void)
{
  return sizeof(UDM_VAR_QUERYWORD);
}


static size_t
UdmValueQueryWordValueSize(void)
{
  return sizeof(UDM_VALUE_QUERYWORD);
}


static size_t
UdmValueQueryWordDataOffset(void)
{
  return offsetof(UDM_VALUE_QUERYWORD, Word);
}


static size_t
UdmValueQueryWordDataSize(void)
{
  return sizeof(UDM_WIDEWORD);
}


static size_t
UdmValueQueryWordDataAlignment(void)
{
  return sizeof(size_t);
}



static udm_rc_t
UdmValueQueryWordDump(const char *Data, const char *name, FILE *f)
{
  return UDM_NOTARGET;
}


static udm_var_flag_t
UdmValueQueryWordFlags(const char *Data)
{
  return UDM_VARFLAG_READONLY;
}


static udm_secno_t
UdmValueQueryWordSecno(const char *Data)
{
  return 0;
}


static size_t
UdmValueQueryWordMemUsed(const char *Data)
{
  const UDM_WIDEWORD *Word= (const UDM_WIDEWORD *) Data;
  return sizeof(UDM_VALUE_QUERYWORD) + Word->Word.length;
}


static void
UdmValueQueryWordGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  const UDM_WIDEWORD *W= (const UDM_WIDEWORD *) Data;
  UdmConstStrSet(Val, W->Word.str, W->Word.length);
}


static void
UdmValueQueryWordGetBool(const char *Data, udm_bool_t *val)
{
  *val= UDM_FALSE;
}


static void
UdmValueQueryWordGetInt(const char *Data, int *val)
{
  *val= 0;
}


static void
UdmValueQueryWordGetUInt(const char *Data, unsigned int *val)
{
  *val= 0;
}


static void
UdmValueQueryWordGetDouble(const char *Data, double *val)
{
  *val= 0;
}


static udm_rc_t
UdmValueQueryWordSetStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueQueryWordSetInt(char *Data, int value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueQueryWordSetDouble(char *Data, double value)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueQueryWordCopy(char *D, const char *S)
{
  UDM_WIDEWORD *dst= (UDM_WIDEWORD *) D;
  const UDM_WIDEWORD *src= (const UDM_WIDEWORD *) S;
  UdmWideWordFree(dst);
  dst->Param= src->Param;
  return UdmStrNDup(&dst->Word, src->Word.str, src->Word.length);
}


static udm_rc_t
UdmValueQueryWordSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                      const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueQueryWordConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueQueryWordAppendStrn(char *Data, const char *str, size_t length)
{
  return UDM_NOTARGET;
}


static udm_rc_t
UdmValueQueryWordAppendConv(char *Data,
                         size_t maxlen,
                         UDM_CONV *cnv, int cnvflags,
                         const char *src, size_t srclen)
{
  return UDM_NOTARGET;
}



static udm_rc_t
UdmValueQueryWordPrintToFile(const char *Data, FILE *f)
{
  UDM_CONST_STR str;
  UdmValueQueryWordGetConstStr(Data, &str);
  fprintf(f, "%.*s", (int) str.length, str.str);
  return UDM_OK;
}


static udm_rc_t
UdmValueQueryWordPrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  UDM_CONST_STR str;
  UdmValueQueryWordGetConstStr(Data, &str);
  UdmDSTRAppend(dstr, str.str, str.length);
  return UDM_OK;
}



/**
  TODO: return address of the left operand.
*/
static void
UdmValueQueryWordOperatorCopy(UDM_PROG_EXECUTOR_STATE *state)
{
  char *dst= state->vreg[0].Data;
  const char *src= state->vreg[1].Data;
  if (dst != src)
    state->fatal_error= UdmValueQueryWordCopy(dst, src);
}


static void
UdmValueQueryWordMethod_count(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_WIDEWORD *Word= (UDM_WIDEWORD *) state->vreg[0].Data; /*this*/
  UDM_FUNC_RETURN_INT(state, (int) Word->Param.count);
}


static void
UdmValueQueryWordMethod_doccount(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_WIDEWORD *Word= (UDM_WIDEWORD *) state->vreg[0].Data; /*this*/
  UDM_FUNC_RETURN_INT(state, (int) Word->Param.doccount);
}


static void
UdmValueQueryWordMethod_order(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_WIDEWORD *Word= (UDM_WIDEWORD *) state->vreg[0].Data; /*this*/
  UDM_FUNC_RETURN_INT(state, (int) Word->Param.order);
}


static void
UdmValueQueryWordMethod_origin(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_WIDEWORD *Word= (UDM_WIDEWORD *) state->vreg[0].Data; /*this*/
  UDM_FUNC_RETURN_INT(state, (int) Word->Param.origin);
}


static void
UdmValueQueryWordMethod_weight(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_WIDEWORD *Word= (UDM_WIDEWORD *) state->vreg[0].Data; /*this*/
  UDM_FUNC_RETURN_INT(state, (int) Word->Param.weight);
}


static void
UdmValueQueryWordMethod_word(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_WIDEWORD *Word= (UDM_WIDEWORD *) state->vreg[0].Data;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UdmDSTRAppend(&ReturnValue->Value,
                                    Word->Word.str, Word->Word.length);
}


static UDM_VALUE_PROTOTYPE UdmValuePrototypeThis[]=
{{&UdmValueHandlerQueryWord}};


static UDM_VALUE_PROTOTYPE UdmValuePrototypeQueryWord2[]=
{
  {&UdmValueHandlerQueryWord},
  {&UdmValueHandlerQueryWord}, 
};


static const UDM_FUNCTION Methods[]=
{
  {{UDM_CSTR_WITH_LEN("operator=")},
    {{NULL}, 2, UdmValuePrototypeQueryWord2},
    UdmValueQueryWordOperatorCopy},

  {{UDM_CSTR_WITH_LEN("count")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmValueQueryWordMethod_count},
  {{UDM_CSTR_WITH_LEN("doccount")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmValueQueryWordMethod_doccount},
  {{UDM_CSTR_WITH_LEN("order")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmValueQueryWordMethod_order},
  {{UDM_CSTR_WITH_LEN("origin")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmValueQueryWordMethod_origin},
  {{UDM_CSTR_WITH_LEN("weight")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeThis},
    UdmValueQueryWordMethod_weight},
  {{UDM_CSTR_WITH_LEN("word")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeThis},
    UdmValueQueryWordMethod_word},
  {{0,0}, {{NULL}, 0, NULL}, NULL}
};


UDM_VALUE_HANDLER UdmValueHandlerQueryWord=
{
  UDM_VALUE_HANDLER_TYPE_STR,
  UDM_VALUE_DATA_TYPE_STR,
  UDM_VALUE_DATA_TYPE_STR,
  "QUERYWORD",
  UdmValueQueryWordConstructor,
  UdmValueQueryWordDestructor,
  UdmValueQueryWordCopy,
  UdmValueQueryWordVarSize,
  UdmValueQueryWordValueSize,
  UdmValueQueryWordDataOffset,
  UdmValueQueryWordDataSize,
  UdmValueQueryWordDataAlignment,
  UdmValueQueryWordDump,
  UdmValueQueryWordFlags,
  UdmValueQueryWordSecno,
  UdmValueQueryWordMemUsed,
  UdmValueQueryWordGetConstStr,
  UdmValueQueryWordGetBool,
  UdmValueQueryWordGetInt,
  UdmValueQueryWordGetUInt,
  UdmValueQueryWordGetDouble,
  UdmValueQueryWordSetInt,
  UdmValueQueryWordSetDouble,
  UdmValueQueryWordSetStrn,
  UdmValueQueryWordPrintToFile,
  UdmValueQueryWordPrintToDSTR,
  UdmValueQueryWordConvertCharset,
  UdmValueQueryWordSetConv,
  UdmValueQueryWordAppendStrn,
  UdmValueQueryWordAppendConv,
  Methods,                 /* Methods */
};
